#include <boost/test/unit_test.hpp>

#include <boost/iterator/widest_convertible.hpp>
#include <boost/iterator/iterator_categories.hpp>
#include <boost/static_assert.hpp>
#include <boost/type_traits.hpp>

BOOST_AUTO_TEST_SUITE(test_widest_convertible)

template<class CatA, class CatB, class CatRes>
void assert_correct_category_symmetric() {
  BOOST_STATIC_ASSERT(( 
     boost::is_same< 
      typename boost::widest_convertible_traversal_cat<CatA, CatB>::type, 
      CatRes 
     >::value 
  ));

  BOOST_STATIC_ASSERT(( 
     boost::is_same< 
      typename boost::widest_convertible_traversal_cat<CatB, CatA>::type, 
      CatRes 
     >::value 
  )); 
};

#define ASSERT_CORRECT_CATEGORY_SYMMETRIC(CatA, CatB, CatRes) \
  assert_correct_category_symmetric< CatA , CatB , CatRes>();


BOOST_AUTO_TEST_CASE(test_incrementable_tag)
{
  ASSERT_CORRECT_CATEGORY_SYMMETRIC(
    boost::incrementable_traversal_tag,
    boost::incrementable_traversal_tag,
    boost::incrementable_traversal_tag
  )

  ASSERT_CORRECT_CATEGORY_SYMMETRIC(
    boost::incrementable_traversal_tag,
    boost::single_pass_traversal_tag,
    boost::incrementable_traversal_tag
  )

  ASSERT_CORRECT_CATEGORY_SYMMETRIC(
    boost::incrementable_traversal_tag,
    boost::forward_traversal_tag,
    boost::incrementable_traversal_tag
  )

  ASSERT_CORRECT_CATEGORY_SYMMETRIC(
    boost::incrementable_traversal_tag,
    boost::bidirectional_traversal_tag,
    boost::incrementable_traversal_tag
  )

  ASSERT_CORRECT_CATEGORY_SYMMETRIC(
    boost::incrementable_traversal_tag,
    boost::random_access_traversal_tag,
    boost::incrementable_traversal_tag
  )
}

BOOST_AUTO_TEST_CASE(test_single_pass_tag)
{
  ASSERT_CORRECT_CATEGORY_SYMMETRIC(
    boost::single_pass_traversal_tag,
    boost::single_pass_traversal_tag,
    boost::single_pass_traversal_tag
  )

  ASSERT_CORRECT_CATEGORY_SYMMETRIC(
    boost::single_pass_traversal_tag,
    boost::forward_traversal_tag,
    boost::single_pass_traversal_tag
  )

  ASSERT_CORRECT_CATEGORY_SYMMETRIC(
    boost::single_pass_traversal_tag,
    boost::bidirectional_traversal_tag,
    boost::single_pass_traversal_tag
  )

  ASSERT_CORRECT_CATEGORY_SYMMETRIC(
    boost::single_pass_traversal_tag,
    boost::random_access_traversal_tag,
    boost::single_pass_traversal_tag
  )
}

BOOST_AUTO_TEST_CASE(test_forward_tag)
{

  ASSERT_CORRECT_CATEGORY_SYMMETRIC(
    boost::forward_traversal_tag,
    boost::forward_traversal_tag,
    boost::forward_traversal_tag
  )

  ASSERT_CORRECT_CATEGORY_SYMMETRIC(
    boost::forward_traversal_tag,
    boost::bidirectional_traversal_tag,
    boost::forward_traversal_tag
  )

  ASSERT_CORRECT_CATEGORY_SYMMETRIC(
    boost::forward_traversal_tag,
    boost::random_access_traversal_tag,
    boost::forward_traversal_tag
  )
}
// copyright christoph.heindl@gmail.com 2009

BOOST_AUTO_TEST_CASE(test_bidir_tag)
{

  ASSERT_CORRECT_CATEGORY_SYMMETRIC(
    boost::bidirectional_traversal_tag,
    boost::bidirectional_traversal_tag,
    boost::bidirectional_traversal_tag
  )

  ASSERT_CORRECT_CATEGORY_SYMMETRIC(
    boost::bidirectional_traversal_tag,
    boost::random_access_traversal_tag,
    boost::bidirectional_traversal_tag
  )
}
  
BOOST_AUTO_TEST_CASE(test_random_tag)
{

  ASSERT_CORRECT_CATEGORY_SYMMETRIC(
    boost::random_access_traversal_tag,
    boost::random_access_traversal_tag,
    boost::random_access_traversal_tag
  )
}

BOOST_AUTO_TEST_SUITE_END()
